from invoke import Collection

from . import (
    benchmark,
    build,
    install,
    lint,
    test,
)

ns = Collection()
# Top level tasks
ns.add_task(benchmark.benchmark)
ns.add_task(build.dist, name="build")
ns.add_task(build.requirements)
ns.add_task(build.release)
ns.add_task(test.test)
# Module collection tasks
ns.add_collection(install)
ns.add_collection(lint)



import datetime
import uuid

import pytest

from syrupy.extensions.amber.serializer import Repr
from syrupy.matchers import (
    PathTypeError,
    path_type,
)


def test_matcher_path_type_noop(snapshot):
    with pytest.raises(PathTypeError, match="argument cannot be empty"):
        path_type()


def test_matches_expected_type(snapshot):
    my_matcher = path_type(
        {"date_created": (datetime.datetime,), "nested.id": (int,)}, types=(uuid.UUID,)
    )
    actual = {
        "date_created": datetime.datetime.now(),
        "nested": {"id": 4},
        "some_uuid": uuid.uuid4(),
    }
    assert actual == snapshot(matcher=my_matcher)


def test_raises_unexpected_type(snapshot):
    kwargs = {
        "mapping": {
            "date_created": (datetime.datetime,),
            "date_updated": (datetime.datetime,),
            "nested.id": (str,),
        },
        "types": (uuid.UUID, int),
    }
    actual = {
        "date_created": datetime.datetime.now(),
        "date_updated": datetime.date(2020, 6, 1),
        "nested": {"id": 4},
        "some_uuid": uuid.uuid4(),
    }
    assert actual == snapshot(matcher=path_type(**kwargs, strict=False))
    with pytest.raises(AssertionError, match="does not match any of the expected"):
        assert actual == snapshot(matcher=path_type(**kwargs))


def test_matches_non_deterministic_snapshots(snapshot):
    def matcher(data, path):
        if isinstance(data, uuid.UUID):
            return Repr("UUID(...)")
        if isinstance(data, datetime.datetime):
            return Repr("datetime.datetime(...)")
        if tuple(p for p, _ in path[-2:]) == ("c", 0):
            return "Your wish is my command"
        return data

    assert {
        "a": uuid.uuid4(),
        "b": {"b_1": "This is deterministic", "b_2": datetime.datetime.now()},
        "c": ["Replace this one", "Do not replace this one"],
    } == snapshot(matcher=matcher)
    assert {
        "a": uuid.UUID("06335e84-2872-4914-8c5d-3ed07d2a2f16"),
        "b": {
            "b_1": "This is deterministic",
            "b_2": datetime.datetime(year=2020, month=5, day=31),
        },
        "c": ["Replace this one", "Do not replace this one"],
    } == snapshot


def test_matches_regex_in_regex_mode(snapshot):
    my_matcher = path_type(
        {
            r"data\.list\..*\.date_created": (datetime.datetime,),
            r"any_number": (int,),
        },
        regex=True,
    )
    actual = {
        "data": {
            "list": [
                {"k": "1", "date_created": datetime.datetime.now()},
                {"k": "2", "date_created": datetime.datetime.now()},
            ],
        },
        "any_number": 3,
        "any_number_adjacent": "hi",
        "specific_number": 5,
    }
    assert actual == snapshot(matcher=my_matcher)



import pytest


@pytest.fixture
def testcases():
    return {
        "base": (
            """
            def test_a(snapshot):
                assert snapshot(name="xyz") == "case 1"
                assert snapshot(name="zyx") == "case 2"
            """
        ),
        "modified": (
            """
            def test_a(snapshot):
                assert snapshot(name="xyz") == "case 1"
                assert snapshot(name="zyx") == "case ??"
            """
        ),
    }


@pytest.fixture
def run_testcases(testdir, testcases):
    testdir.makepyfile(test_1=testcases["base"])
    result = testdir.runpytest(
        "-v",
        "--snapshot-update",
    )
    result.stdout.re_match_lines((r"2 snapshots generated\."))
    return testdir, testcases


def test_run_all(run_testcases):
    testdir, testcases = run_testcases
    result = testdir.runpytest(
        "-v",
    )
    result.stdout.re_match_lines("2 snapshots passed")
    assert result.ret == 0


def test_failure(run_testcases):
    testdir, testcases = run_testcases
    testdir.makepyfile(test_1=testcases["modified"])
    result = testdir.runpytest(
        "-v",
    )
    result.stdout.re_match_lines("1 snapshot failed. 1 snapshot passed.")
    assert result.ret == 1


def test_update(run_testcases):
    testdir, testcases = run_testcases
    testdir.makepyfile(test_1=testcases["modified"])
    result = testdir.runpytest(
        "-v",
        "--snapshot-update",
    )
    assert "Can not relate snapshot name" not in str(result.stdout)
    result.stdout.re_match_lines("1 snapshot passed. 1 snapshot updated.")
    assert result.ret == 0



import pytest


@pytest.fixture
def testcases():
    return {
        "used": (
            """
            def test_used(snapshot):
                assert snapshot == 'used'
            """
        ),
        "unused": (
            """
            def test_unused(snapshot):
                assert snapshot == 'unused'
            """
        ),
    }


@pytest.fixture
def extra_testcases():
    return {
        "extra_a": (
            """
            def test_extra_a(snapshot):
                assert snapshot == 'extra_a'
            """
        ),
        "extra_b": (
            """
            def test_extra_b(snapshot):
                assert snapshot == 'extra_b'
            """
        ),
    }


@pytest.fixture
def run_testfiles_with_update(testdir):
    def run_testfiles_with_update_impl(**testfiles):
        testdir.makepyfile(
            **{
                filename: "\n\n".join(cases.values())
                for (filename, cases) in testfiles.items()
            }
        )
        result = testdir.runpytest("-v", "--snapshot-update")
        result.stdout.re_match_lines((r"[0-9]+ snapshots generated\."))
        return testdir

    return run_testfiles_with_update_impl


@pytest.mark.parametrize(
    (
        "options",
        "expected_status_code",
    ),
    (
        (("-v", "--snapshot-details"), 1),
        (("-v", "--snapshot-details", "--snapshot-warn-unused"), 0),
    ),
)
def test_unused_snapshots_details(
    options, expected_status_code, run_testfiles_with_update, testcases
):
    testdir = run_testfiles_with_update(test_file=testcases)
    testdir.makepyfile(test_file=testcases["used"])

    result = testdir.runpytest(*options)
    result.stdout.re_match_lines(
        (
            r"1 snapshot passed\. 1 snapshot unused\.",
            r"Unused test_unused \(__snapshots__[\\/]test_file.ambr\)",
            r"Re-run pytest with --snapshot-update to delete unused snapshots\.",
        )
    )
    assert result.ret == expected_status_code


def test_unused_snapshots_details_multiple_tests(
    run_testfiles_with_update, testcases, extra_testcases
):
    testdir = run_testfiles_with_update(
        test_file=testcases, test_second_file=extra_testcases
    )
    testdir.makepyfile(
        test_file="\n\n".join(testcases.values()),
        test_second_file="",
    )

    result = testdir.runpytest("-v", "--snapshot-details")
    result.stdout.re_match_lines(
        (
            r"2 snapshots passed\. 2 snapshots unused\.",
            r"Unused test_extra_a, test_extra_b "
            r"\(__snapshots__[\\/]test_second_file.ambr\)",
            r"Re-run pytest with --snapshot-update to delete unused snapshots\.",
        )
    )
    assert result.ret == 1


def test_unused_snapshots_details_multiple_locations(
    run_testfiles_with_update, testcases, extra_testcases
):
    testdir = run_testfiles_with_update(
        test_file=testcases, test_second_file=extra_testcases
    )
    testdir.makepyfile(
        test_file=testcases["used"],
        test_second_file=extra_testcases["extra_a"],
    )

    result = testdir.runpytest("-v", "--snapshot-details")
    result.stdout.re_match_lines_random(
        (
            r"2 snapshots passed\. 2 snapshots unused\.",
            r"Unused test_extra_b \(__snapshots__[\\/]test_second_file.ambr\)",
            r"Unused test_unused \(__snapshots__[\\/]test_file.ambr\)",
            r"Re-run pytest with --snapshot-update to delete unused snapshots\.",
        )
    )
    assert result.ret == 1


def test_unused_snapshots_details_no_details_on_deletion(
    run_testfiles_with_update, testcases
):
    testdir = run_testfiles_with_update(test_file=testcases)
    testdir.makepyfile(test_file=testcases["used"])

    result = testdir.runpytest("-v", "--snapshot-details", "--snapshot-update")
    result.stdout.re_match_lines(
        (
            r"1 snapshot passed\. 1 unused snapshot deleted\.",
            r"Deleted test_unused \(__snapshots__[\\/]test_file.ambr\)",
        )
    )
    assert result.ret == 0



import os

from invoke import task

from .utils import ctx_run


@task(
    help={
        "coverage": "Build and report on test coverage",
        "test-pattern": "Pattern used to select test files to run",
        "update-snapshots": "Create, update or delete snapshot files",
        "verbose": "Verbose output e.g. non captured logs etc.",
    }
)
def test(
    ctx,
    coverage=False,
    test_pattern=None,
    update_snapshots=False,
    verbose=False,
    debug=False,
):
    """
    Run entire test suite
    """
    flags = {
        "-s -vv": verbose,
        f"-k {test_pattern}": test_pattern,
        "--snapshot-update": update_snapshots,
        "--pdb": debug,
    }
    coverage_module = "coverage run -m " if coverage else ""
    test_flags = " ".join(flag for flag, enabled in flags.items() if enabled)
    ctx_run(ctx, f"python -m {coverage_module}pytest {test_flags} .")
    if coverage:
        if not os.environ.get("CI") or not os.environ.get("CODECOV_TOKEN"):
            ctx_run(ctx, "coverage report")
        else:
            ctx_run(ctx, "codecov")



import pytest


@pytest.fixture
def testcases_initial(testdir):
    testdir.makeconftest(
        """
        import pytest

        from syrupy.extensions.amber import AmberSnapshotExtension
        from syrupy.extensions.image import (
            PNGImageSnapshotExtension,
            SVGImageSnapshotExtension,
        )
        from syrupy.extensions.single_file import SingleFileSnapshotExtension


        class CustomSnapshotExtension(AmberSnapshotExtension):
            @property
            def _file_extension(self):
                return ""

            def serialize(self, data, **kwargs):
                return str(data)

            def get_snapshot_name(self, *, index):
                testname = self._test_location.testname[::-1]
                return f"{testname}.{index}"

            def _get_file_basename(self, *, index):
                return self.test_location.filename[::-1]

        @pytest.fixture
        def snapshot_custom(snapshot):
            return snapshot.use_extension(CustomSnapshotExtension)


        @pytest.fixture
        def snapshot_single(snapshot):
            return snapshot.use_extension(SingleFileSnapshotExtension)


        @pytest.fixture
        def snapshot_png(snapshot):
            return snapshot.use_extension(PNGImageSnapshotExtension)


        @pytest.fixture
        def snapshot_svg(snapshot):
            return snapshot.use_extension(SVGImageSnapshotExtension)
        """
    )
    return {
        "passed": (
            """
            def test_passed_custom(snapshot_custom):
                assert snapshot_custom == 'passed1'
                assert snapshot_custom == 'passed2'

            def test_passed_single(snapshot_single):
                assert snapshot_single == b'passed1'
                assert snapshot_single == b'passed2'
            """
        ),
        "failed": (
            """
            def test_failed_single(snapshot_single):
                assert snapshot_single == 'failed'

            def test_failed_image(snapshot_png):
                assert "not a byte string" == snapshot_png
            """
        ),
    }


@pytest.fixture
def testcases_updated(testcases_initial):
    updated_testcases = {
        "passed": (
            """
            def test_passed_single(snapshot_single):
                assert snapshot_single == b'passed'
            """
        )
    }
    return {**testcases_initial, **updated_testcases}


@pytest.fixture
def generate_snapshots(testdir, testcases_initial):
    testdir.makepyfile(test_file=testcases_initial["passed"])
    result = testdir.runpytest("-v", "--snapshot-update")
    return result, testdir, testcases_initial


def test_unsaved_snapshots(testdir, testcases_initial):
    testdir.makepyfile(test_file=testcases_initial["passed"])
    result = testdir.runpytest("-v")
    result.stdout.re_match_lines(
        (r".*Snapshot 'test_passed_single' does not exist!", r".*\+ b'passed1'")
    )
    assert result.ret == 1


def test_failed_snapshots(testdir, testcases_initial):
    testdir.makepyfile(test_file=testcases_initial["failed"])
    result = testdir.runpytest("-v", "--snapshot-update")
    result.stdout.re_match_lines((r"2 snapshots failed\."))
    assert result.ret == 1


def test_generated_snapshots(generate_snapshots):
    result = generate_snapshots[0]
    result.stdout.re_match_lines((r"4 snapshots generated\."))
    assert "snapshots unused" not in result.stdout.str()
    assert result.ret == 0


def test_unmatched_snapshots(generate_snapshots, testcases_updated):
    testdir = generate_snapshots[1]
    testdir.makepyfile(test_file=testcases_updated["passed"])
    result = testdir.runpytest("-v")
    result.stdout.re_match_lines((r"1 snapshot failed\. 2 snapshots unused\."))
    assert result.ret == 1


def test_updated_snapshots(generate_snapshots, testcases_updated):
    testdir = generate_snapshots[1]
    testdir.makepyfile(test_file=testcases_updated["passed"])
    result = testdir.runpytest("-v", "--snapshot-update")
    result.stdout.re_match_lines((r"1 snapshot updated\. 2 unused snapshots deleted\."))
    assert result.ret == 0


def test_warns_on_snapshot_name(generate_snapshots):
    result = generate_snapshots[0]
    result.stdout.re_match_lines(
        (
            r".*Warning:\s+",
            r"\s+Can not relate snapshot location",
            r"\s+Can not relate snapshot name",
            r"4 snapshots generated\.",
        )
    )
    assert result.ret == 0



import os
import subprocess
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    Tuple,
)

from github import (
    Github,
    GithubException,
    UnknownObjectException,
)

if TYPE_CHECKING:
    from subprocess import CompletedProcess

    from github.Commit import Commit

BENCH_COMMAND = "pytest benchmarks/test.py -qq --snapshot-update"
BENCH_PERF_FILE = "current.json"
BENCH_REF_FILE = "master.json"
GH_BENCH_FILE_PATH = "runs"
GH_BENCH_BRANCH = "benchmarks"
GH_BRANCH_REF = "master"
GH_REPO = "tophat/syrupy"
GH_STATUS_CONTEXT = "Syrupy CICD / Benchmark"
GH_STATUS_DESC_DELIM = "\n" * 3


def get_req_env(var_name: str) -> str:
    """
    Try to get environment variable and exits if not available
    """
    try:
        return os.environ[var_name]
    except KeyError:
        print(f"Missing required environment variable '{var_name}'.")
        exit(1)


def get_commit(github: "Github") -> "Commit":
    """
    Get current commit the github ci run is for
    """
    commit_sha = get_req_env("COMMIT_SHA")
    return github.get_repo(GH_REPO).get_commit(commit_sha)


def get_branch() -> Optional[str]:
    """
    Get current branch the github ci run is for
    """
    github_ref = get_req_env("GITHUB_REF")
    return github_ref.replace("refs/heads/", "") if github_ref else None


def get_target_url() -> str:
    """
    Get url to benchmark result
    """
    repo_url = f"https://github.com/{GH_REPO}"
    all_results_url = f"{repo_url}/tree/{GH_BENCH_BRANCH}/{GH_BENCH_FILE_PATH}"
    commit_sha = get_req_env("COMMIT_SHA")
    return all_results_url + (f"/{commit_sha}" if commit_sha else "")


def report_pending(github: Optional["Github"] = None) -> None:
    """
    Set the commit benchmark status to pending
    """
    if not github:
        return
    get_commit(github).create_status(
        state="pending", description="Running benchmarks", context=GH_STATUS_CONTEXT
    )


def default_runner(cmd: str, **kwargs: Any) -> "CompletedProcess[bytes]":
    """
    Default code runner, interchangeable with invoke context runner
    """
    return subprocess.run(cmd, shell=True, check=True, **kwargs)


def get_commit_bench_path(commit_sha: str) -> str:
    return f"{GH_BENCH_FILE_PATH}/{commit_sha}/result.json"


def measure_perf(
    github: Optional["Github"], run: Callable[..., Any] = default_runner
) -> None:
    """
    Measure benchmark command performance and save results in file
    Checks that the benchmark run was stable
    Saves the benchmark results to github
    """
    run(f"python -m pyperf command -o {BENCH_PERF_FILE} -- {BENCH_COMMAND}")
    run(f"python -m pyperf check {BENCH_PERF_FILE}")
    if not github:
        return
    repo = github.get_repo(GH_REPO)
    commit_sha = get_req_env("COMMIT_SHA")
    with open(BENCH_PERF_FILE, "r", encoding="utf-8") as bench_file:
        try:
            repo.create_file(
                path=get_commit_bench_path(commit_sha),
                message=f"build: benchmark run {commit_sha[:7]}",
                content=bench_file.read(),
                branch=GH_BENCH_BRANCH,
            )
        except GithubException:
            print("Unable to save benchmark results to repo")


def fetch_branch_bench_json(github: "Github", branch: str) -> Optional[str]:
    """
    Retrieve the benchmark results for the head commit of the given branch
    """
    repo = github.get_repo(GH_REPO)
    commit_sha = repo.get_branch(branch).commit.sha
    commit_bench_path = get_commit_bench_path(commit_sha)
    try:
        content_file = repo.get_contents(commit_bench_path, GH_BENCH_BRANCH)
        if isinstance(content_file, list):
            raise UnknownObjectException
        return str(content_file.decoded_content.decode())
    except UnknownObjectException:
        print("Unable to retrieve benchmark results from repo")
        return None


def fetch_ref_bench_json(github: "Github", ref_branch: str = GH_BRANCH_REF) -> bool:
    """
    Retrieve and save the benchmark results on the reference branch
    Skipping the current branch is the reference branch
    """
    if get_branch() == ref_branch:
        return False

    ref_json = fetch_branch_bench_json(github, ref_branch)
    if not ref_json:
        return False

    with open(BENCH_REF_FILE, "w", encoding="utf-8") as ref_json_file:
        ref_json_file.write(ref_json)
    return True


def compare_bench_status(
    perf_file: str = BENCH_PERF_FILE, ref_file: str = BENCH_REF_FILE
) -> Tuple[str, bool]:
    """
    Compare benchmarks for current branch to reference and return results
    indicating success or failure and the github status description
    """
    cmd = f"python -m pyperf compare_to -q {ref_file} {perf_file}"
    return default_runner(cmd, stdout=subprocess.PIPE).stdout.decode().strip(), True


def get_bench_status(bench_file: str = BENCH_PERF_FILE) -> Tuple[str, bool]:
    """
    Get the status of the benchmark file indicating success or failure
    and the github status description
    """
    cmd = f"python -m pyperf show -q {bench_file}"
    return default_runner(cmd, stdout=subprocess.PIPE).stdout.decode().strip(), True


def report_status(github: Optional["Github"] = None) -> None:
    """
    Report benchmark run results to github as the commit status
    """
    if not github:
        return

    if fetch_ref_bench_json(github):
        description, success = compare_bench_status()
    else:
        description, success = get_bench_status()

    get_commit(github).create_status(
        state="success" if success else "failure",
        target_url=get_target_url(),
        description=description,
        context=GH_STATUS_CONTEXT,
    )


def report_error(github: Optional["Github"] = None) -> None:
    """
    Report benchmark error to github as the commit status
    """
    if not github:
        return
    get_commit(github).create_status(
        state="error",
        description="Unable to report benchmark",
        context=GH_STATUS_CONTEXT,
    )


def main(report: bool = False) -> None:
    github = Github(get_req_env("GH_TOKEN")) if report else None
    try:
        report_pending(github)
        measure_perf(github)
        report_status(github)
    except Exception:
        report_error(github)



from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from syrupy.data import (
    Snapshot,
    SnapshotFossil,
)
from syrupy.extensions.single_file import SingleFileSnapshotExtension

if TYPE_CHECKING:
    from syrupy.assertion import SnapshotAssertion


@pytest.fixture
def snapshot_single(snapshot):
    return snapshot.use_extension(SingleFileSnapshotExtension)


class SingleFileUTF8SnapshotExtension(SingleFileSnapshotExtension):
    def serialize(self, data, **kwargs) -> bytes:
        return bytes(data, "utf8")


@pytest.fixture
def snapshot_utf8(snapshot):
    return snapshot.use_extension(SingleFileUTF8SnapshotExtension)


def test_does_not_write_non_binary(testdir, snapshot_single: "SnapshotAssertion"):
    snapshot_fossil = SnapshotFossil(
        location=str(Path(testdir.tmpdir).joinpath("snapshot_fossil.raw")),
    )
    snapshot_fossil.add(Snapshot(name="snapshot_name", data="non binary data"))
    with pytest.raises(TypeError, match="Expected 'bytes', got 'str'"):
        snapshot_single.extension._write_snapshot_fossil(
            snapshot_fossil=snapshot_fossil
        )
    assert not Path(snapshot_fossil.location).exists()


class TestClass:
    def test_class_method_name(self, snapshot_single):
        assert snapshot_single == b"this is in a test class"

    @pytest.mark.parametrize("content", [b"x", b"y", b"z"])
    def test_class_method_parametrized(self, snapshot_single, content):
        assert snapshot_single == content


def test_underscore(snapshot_single):
    assert snapshot_single == b"apple"


def test_____underscore(snapshot_single):
    assert snapshot_single == b"orange"


@pytest.mark.parametrize(
    "content", [b"", b"_", b"a?", b"space space", b".123~!@#$%^&*()/[]{}|"]
)
def test_special_characters(snapshot_single, content):
    assert snapshot_single == content


@pytest.mark.parametrize("content", ["greek ῴ"])
def test_unicode(snapshot_utf8, content):
    assert snapshot_utf8 == "apple"



import pytest


@pytest.fixture
def testcases():
    return {
        "inject": (
            """
            def test_injection(snapshot):
                assert snapshot is not None
            """
        ),
    }


@pytest.fixture
def run_testcases(testdir, testcases):
    pyfile_content = "\n\n".join(testcases.values())
    testdir.makepyfile(test_file=pyfile_content)
    return testdir.runpytest("-v")


def test_injected_fixture(run_testcases):
    run_testcases.stdout.fnmatch_lines(["*::test_injection PASSED*"])
    assert run_testcases.ret == 0



import pytest

pytest_version = tuple(int(v) for v in pytest.__version__.split("."))


@pytest.mark.skipif(pytest_version < (6, 0), reason="at least pytest 6 required")
def test_ignores_non_function_nodes(testdir):
    conftest = """
        import pytest

        class CustomItem(pytest.Item, pytest.File):
            def __init__(self, *args, fspath, parent, **kwargs):
                super().__init__(fspath, parent=parent)
                self._nodeid += "::CUSTOM"

            def runtest(self):
                pass

        def pytest_collect_file(path, parent):
            return CustomItem.from_parent(fspath=path, parent=parent)
        """
    testcase = """
        def test_example(snapshot):
            assert snapshot == 1
        """
    testdir.makeconftest(conftest)
    testdir.makepyfile(test_file=testcase)
    result = testdir.runpytest("test_file.py", "-v", "--snapshot-update")
    result.stdout.re_match_lines((r".*test_file.py::CUSTOM.*"))
    assert result.ret == 0


def test_handles_pyargs_non_module_when_both_given(testdir):
    testdir.makeconftest("")
    testcase = """
        def test_example(snapshot):
            assert snapshot == 1
        """
    testdir.makepyfile(test_file=testcase)
    result = testdir.runpytest(
        "-v", "test_file.py", "--pyargs", "test_file", "--snapshot-update"
    )
    assert result.ret == 0



import os
import re

from invoke import task

from setup import install_requires

from .utils import ctx_run


def _parse_min_versions(requirements):
    result = []
    for req in sorted(requirements):
        match = re.match(r"([\w_]+)>=([^,]+),.*", req)
        if match is None:
            continue
        pkg_name = match.group(1)
        min_version = match.group(2)
        result.append(f"{pkg_name}=={min_version}")
    return result


@task
def requirements(ctx, upgrade=False):
    """
    Build test & dev requirements lock file
    """
    args = [
        "--no-emit-find-links",
        "--no-emit-index-url",
        "--allow-unsafe",
        "--rebuild",
    ]
    if upgrade:
        args.append("--upgrade")
    ctx_run(
        ctx,
        f"echo '-e .[dev]' | python -m piptools compile "
        f"{' '.join(args)} - -qo- | sed '/^-e / d' > dev_requirements.txt",
    )

    with open("min_requirements.constraints", "w", encoding="utf-8") as f:
        min_requirements = _parse_min_versions(install_requires)
        f.write("\n".join(min_requirements))
        f.write("\n")


@task
def clean(ctx):
    """
    Remove build files e.g. package, distributable, compiled etc.
    """
    ctx_run(ctx, "rm -rf *.egg-info dist build __pycache__ .pytest_cache artifacts/*")


@task(pre=[clean])
def dist(ctx):
    """
    Generate version from scm and build package distributable
    """
    ctx_run(ctx, "python setup.py sdist bdist_wheel")


@task
def publish(ctx, dry_run=True):
    """
    Upload built package to pypi
    """
    repo_url = "--repository-url https://test.pypi.org/legacy/" if dry_run else ""
    ctx_run(ctx, f"twine upload --skip-existing {repo_url} dist/*")


@task(pre=[dist])
def release(ctx, dry_run=True):
    """
    Build and publish package to pypi index based on scm version
    """
    from semver import parse_version_info

    if not dry_run and not os.environ.get("CI"):
        print("This is a CI only command")
        exit(1)

    # get version created in build
    with open("version.txt", "r", encoding="utf-8") as f:
        version = str(f.read())

    try:
        should_publish_to_pypi = not dry_run and parse_version_info(version)
    except ValueError:
        should_publish_to_pypi = False

    # publish to test to verify builds
    if dry_run:
        publish(ctx, dry_run=True)

    # publish to pypi if test succeeds
    if should_publish_to_pypi:
        publish(ctx, dry_run=False)



import datetime

import pytest

from syrupy.filters import (
    paths,
    props,
)


def test_filters_path_noop():
    with pytest.raises(TypeError, match="required positional argument"):
        paths()


def test_filters_expected_paths(snapshot):
    actual = {
        0: "some value",
        "date": datetime.datetime.now(),
        "nested": {"id": 4, "other": "value"},
        "list": [1, 2],
    }
    assert actual == snapshot(exclude=paths("0", "date", "nested.id", "list.0"))


def test_filters_prop_noop():
    with pytest.raises(TypeError, match="required positional argument"):
        props()


def test_filters_expected_props(snapshot):
    actual = {
        0: "some value",
        "date": datetime.datetime.now(),
        "nested": {"id": 4, "other": "value"},
        "list": [1, 2],
    }
    assert actual == snapshot(exclude=props("0", "date", "id"))


@pytest.mark.parametrize(
    "predicate", [paths("exclude_me", "nested.exclude_me"), props("exclude_me")]
)
def test_filters_error_prop(snapshot, predicate):
    class CustomClass:
        @property
        def include_me(self):
            return "prop value"

        @property
        def exclude_me(self):
            raise Exception("Why you no exclude me?")

    class WithNested(CustomClass):
        nested = CustomClass()

    assert WithNested() == snapshot(exclude=predicate)



import traceback
from gettext import gettext
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Type,
    Union,
)

import attr

from .exceptions import SnapshotDoesNotExist

if TYPE_CHECKING:
    from .extensions.base import AbstractSyrupyExtension
    from .location import PyTestLocation
    from .session import SnapshotSession
    from .types import (
        PropertyFilter,
        PropertyMatcher,
        SerializableData,
        SerializedData,
    )


@attr.s
class AssertionResult:
    snapshot_location: str = attr.ib()
    snapshot_name: str = attr.ib()
    asserted_data: Optional["SerializedData"] = attr.ib()
    recalled_data: Optional["SerializedData"] = attr.ib()
    created: bool = attr.ib()
    updated: bool = attr.ib()
    success: bool = attr.ib()
    exception: Optional[Exception] = attr.ib()

    @property
    def final_data(self) -> Optional["SerializedData"]:
        if self.created or self.updated:
            return self.asserted_data
        return self.recalled_data


@attr.s(cmp=False, repr=False)
class SnapshotAssertion:
    name: str = attr.ib(default="snapshot")
    _session: "SnapshotSession" = attr.ib(kw_only=True)
    _extension_class: Type["AbstractSyrupyExtension"] = attr.ib(kw_only=True)
    _test_location: "PyTestLocation" = attr.ib(kw_only=True)
    _update_snapshots: bool = attr.ib(kw_only=True)
    _exclude: Optional["PropertyFilter"] = attr.ib(
        init=False, default=None, kw_only=True
    )
    _custom_index: Optional[str] = attr.ib(init=False, default=None, kw_only=True)
    _extension: Optional["AbstractSyrupyExtension"] = attr.ib(
        init=False, default=None, kw_only=True
    )
    _executions: int = attr.ib(init=False, default=0, kw_only=True)
    _execution_results: Dict[int, "AssertionResult"] = attr.ib(
        init=False, factory=dict, kw_only=True
    )
    _matcher: Optional["PropertyMatcher"] = attr.ib(
        init=False, default=None, kw_only=True
    )
    _post_assert_actions: List[Callable[..., None]] = attr.ib(
        init=False, factory=list, kw_only=True
    )

    def __attrs_post_init__(self) -> None:
        self._session.register_request(self)

    def __init_extension(
        self, extension_class: Type["AbstractSyrupyExtension"]
    ) -> "AbstractSyrupyExtension":
        return extension_class(test_location=self._test_location)

    @property
    def extension(self) -> "AbstractSyrupyExtension":
        if not self._extension:
            self._extension = self.__init_extension(self._extension_class)
        return self._extension

    @property
    def num_executions(self) -> int:
        return int(self._executions)

    @property
    def executions(self) -> Dict[int, AssertionResult]:
        return self._execution_results

    @property
    def index(self) -> Union[str, int]:
        if self._custom_index:
            return self._custom_index
        return self.num_executions

    def use_extension(
        self, extension_class: Optional[Type["AbstractSyrupyExtension"]] = None
    ) -> "SnapshotAssertion":
        """
        Creates a new snapshot assertion fixture with the same options but using
        specified extension class. This does not preserve assertion index or state.
        """
        return self.__class__(
            update_snapshots=self._update_snapshots,
            test_location=self._test_location,
            extension_class=extension_class or self._extension_class,
            session=self._session,
        )

    def assert_match(self, data: "SerializableData") -> None:
        assert self == data

    def _serialize(self, data: "SerializableData") -> "SerializedData":
        return self.extension.serialize(
            data, exclude=self._exclude, matcher=self._matcher
        )

    def get_assert_diff(self) -> List[str]:
        assertion_result = self._execution_results[self.num_executions - 1]
        if assertion_result.exception:
            lines = [
                line
                for lines in traceback.format_exception(
                    assertion_result.exception.__class__,
                    assertion_result.exception,
                    assertion_result.exception.__traceback__,
                )
                for line in lines.splitlines()
            ]
            # Rotate to place exception with message at first line
            return lines[-1:] + lines[:-1]
        snapshot_data = assertion_result.recalled_data
        serialized_data = assertion_result.asserted_data or ""
        diff: List[str] = []
        if snapshot_data is None:
            diff.append(
                gettext("Snapshot '{}' does not exist!").format(
                    assertion_result.snapshot_name
                )
            )
        if not assertion_result.success:
            diff.extend(self.extension.diff_lines(serialized_data, snapshot_data or ""))
        return diff

    def __with_prop(self, prop_name: str, prop_value: Any) -> None:
        setattr(self, prop_name, prop_value)
        self._post_assert_actions.append(lambda: setattr(self, prop_name, None))

    def __call__(
        self,
        *,
        exclude: Optional["PropertyFilter"] = None,
        extension_class: Optional[Type["AbstractSyrupyExtension"]] = None,
        matcher: Optional["PropertyMatcher"] = None,
        name: Optional[str] = None,
    ) -> "SnapshotAssertion":
        """
        Modifies assertion instance options
        """
        if exclude:
            self.__with_prop("_exclude", exclude)
        if extension_class:
            self.__with_prop("_extension", self.__init_extension(extension_class))
        if matcher:
            self.__with_prop("_matcher", matcher)
        if name:
            self.__with_prop("_custom_index", name)
        return self

    def __dir__(self) -> List[str]:
        return ["name", "num_executions"]

    def __eq__(self, other: "SerializableData") -> bool:
        return self._assert(other)

    def _assert(self, data: "SerializableData") -> bool:
        snapshot_location = self.extension.get_location(index=self.index)
        snapshot_name = self.extension.get_snapshot_name(index=self.index)
        snapshot_data: Optional["SerializedData"] = None
        serialized_data: Optional["SerializedData"] = None
        matches = False
        assertion_success = False
        assertion_exception = None
        try:
            snapshot_data = self._recall_data()
            serialized_data = self._serialize(data)
            matches = snapshot_data is not None and serialized_data == snapshot_data
            assertion_success = matches
            if not matches and self._update_snapshots:
                self.extension.write_snapshot(
                    data=serialized_data,
                    index=self.index,
                )
                assertion_success = True
            return assertion_success
        except Exception as e:
            assertion_exception = e
            return False
        finally:
            snapshot_created = snapshot_data is None and assertion_success
            snapshot_updated = matches is False and assertion_success
            self._execution_results[self._executions] = AssertionResult(
                snapshot_location=snapshot_location,
                snapshot_name=snapshot_name,
                recalled_data=snapshot_data,
                asserted_data=serialized_data,
                success=assertion_success,
                created=snapshot_created,
                updated=snapshot_updated,
                exception=assertion_exception,
            )
            self._executions += 1
            self._post_assert()

    def _post_assert(self) -> None:
        """
        Restores assertion instance options
        """
        while self._post_assert_actions:
            self._post_assert_actions.pop()()

    def _recall_data(self) -> Optional["SerializableData"]:
        try:
            return self.extension.read_snapshot(index=self.index)
        except SnapshotDoesNotExist:
            return None



from invoke import task


@task(help={"report": "Publish report as github status"})
def benchmark(ctx, report=False):
    """
    Run and generate benchmarks for current code
    """
    import benchmarks

    benchmarks.main(report=report)



import base64

import pytest

from syrupy.extensions.image import PNGImageSnapshotExtension


@pytest.fixture
def snapshot(snapshot):
    return snapshot.use_extension(PNGImageSnapshotExtension)


def test_png_image_with_custom_name_suffix(snapshot):
    reddish_square = base64.b64decode(
        b"iVBORw0KGgoAAAANSUhEUgAAAAQAAAAECAIAAAAmkwkpAAAAIUlEQVQIHTXB"
        b"MQEAAAABQUYtvpD+dUzu3KBzg84NOjfoBjmmAd3WpSsrAAAAAElFTkSuQmCC"
    )

    blueish_square = base64.b64decode(
        b"iVBORw0KGgoAAAANSUhEUgAAAAQAAAAECAIAAAAmkwkpAAAAIUlEQVQIHTXB"
        b"MQEAAAABQUYtvpD+dUzuTKozqc6kOpPqBjg+Ad2g/BLMAAAAAElFTkSuQmCC"
    )

    assert blueish_square == snapshot(name="blueish")
    assert reddish_square == snapshot(name="reddish")



from typing import (
    Any,
    Callable,
    TypeVar,
)

ReturnType = TypeVar("ReturnType")

def fixture(func: Callable[..., ReturnType]) -> Callable[..., ReturnType]: ...

class Function: ...
class Item: ...



from typing import (
    Any,
    Union,
)

import colored

from .constants import DISABLE_COLOR_ENV_VARS
from .utils import get_env_value


def _is_color_disabled() -> bool:
    return any(map(get_env_value, DISABLE_COLOR_ENV_VARS))


def _stylize(text: Union[str, int], *args: Any) -> str:
    if _is_color_disabled():
        return str(text)
    return colored.stylize(text, *args)


def reset(text: Union[str, int]) -> str:
    return _stylize(text, colored.attr("reset"))


def red(text: Union[str, int]) -> str:
    return _stylize(text, colored.fg("red"))


def yellow(text: Union[str, int]) -> str:
    return _stylize(text, colored.fg("yellow"))


def green(text: Union[str, int]) -> str:
    return _stylize(text, colored.fg("green"))


def bold(text: Union[str, int]) -> str:
    return _stylize(text, colored.attr("bold"))


def error_style(text: Union[str, int]) -> str:
    return bold(red(text))


def warning_style(text: Union[str, int]) -> str:
    return bold(yellow(text))


def success_style(text: Union[str, int]) -> str:
    return bold(green(text))


def snapshot_style(text: Union[str, int]) -> str:
    return _stylize(text, colored.bg(225) + colored.fg(90))


def snapshot_diff_style(text: Union[str, int]) -> str:
    return _stylize(text, colored.bg(90) + colored.fg(225))


def received_style(text: Union[str, int]) -> str:
    return _stylize(text, colored.bg(195) + colored.fg(23))


def received_diff_style(text: Union[str, int]) -> str:
    return _stylize(text, colored.bg(23) + colored.fg(195))


def context_style(text: Union[str, int]) -> str:
    return _stylize(text, colored.attr("dim"))



import pytest


@pytest.fixture
def testcases():
    return {
        "a": (
            """
            def test_a(snapshot):
                assert snapshot == b"a"
            """
        ),
        "b": (
            """
            def test_b(snapshot):
                assert snapshot == b"b"
            """
        ),
    }


@pytest.fixture
def run_testcases(testdir, testcases):
    pyfile_content = "\n\n".join(testcases.values())
    testdir.makepyfile(test_1=pyfile_content, test_2=pyfile_content)
    result = testdir.runpytest(
        "-v",
        "--snapshot-update",
        "--snapshot-default-extension",
        "syrupy.extensions.single_file.SingleFileSnapshotExtension",
    )
    result.stdout.re_match_lines((r"4 snapshots generated\."))
    return testdir, testcases


def test_run_all(run_testcases):
    testdir, testcases = run_testcases
    result = testdir.runpytest(
        "-v",
        "--snapshot-default-extension",
        "syrupy.extensions.single_file.SingleFileSnapshotExtension",
    )
    result.stdout.re_match_lines("4 snapshots passed")
    assert result.ret == 0


def test_run_single_file(run_testcases):
    testdir, testcases = run_testcases
    result = testdir.runpytest(
        "-v",
        "--snapshot-default-extension",
        "syrupy.extensions.single_file.SingleFileSnapshotExtension",
        "test_1.py",
    )
    result.stdout.re_match_lines("2 snapshots passed")
    assert result.ret == 0


def test_run_single_test_case_in_file(run_testcases):
    testdir, testcases = run_testcases
    result = testdir.runpytest(
        "-v",
        "--snapshot-default-extension",
        "syrupy.extensions.single_file.SingleFileSnapshotExtension",
        "test_2.py::test_a",
    )
    result.stdout.re_match_lines("1 snapshot passed")
    assert result.ret == 0


def test_run_all_but_one(run_testcases):
    testdir, testcases = run_testcases
    result = testdir.runpytest(
        "-v",
        "--snapshot-details",
        "--snapshot-default-extension",
        "syrupy.extensions.single_file.SingleFileSnapshotExtension",
        "test_1.py",
        "test_2.py::test_a",
    )
    result.stdout.re_match_lines("3 snapshots passed")
    assert result.ret == 0


def test_run_both_files_by_node(run_testcases):
    testdir, testcases = run_testcases
    result = testdir.runpytest(
        "-v",
        "--snapshot-details",
        "--snapshot-default-extension",
        "syrupy.extensions.single_file.SingleFileSnapshotExtension",
        "test_1.py::test_a",
        "test_2.py::test_a",
    )
    result.stdout.re_match_lines("2 snapshots passed")
    assert result.ret == 0


def test_run_both_files_by_node_2(run_testcases):
    testdir, testcases = run_testcases
    result = testdir.runpytest(
        "-v",
        "--snapshot-details",
        "--snapshot-default-extension",
        "syrupy.extensions.single_file.SingleFileSnapshotExtension",
        "test_1.py::test_b",
        "test_2.py::test_a",
    )
    result.stdout.re_match_lines("2 snapshots passed")
    assert result.ret == 0



"""
Example: Custom Snapshot Directory

Here we extend the Amber extension to change the directory
in which snapshots are stored.

We explicitly name our new fixture "snapshot" to override the
default snapshot fixture. If this is placed in your project's
root conftest.py file, it is equivalent to globally overriding
the default snapshot directory.
"""

from pathlib import Path

import pytest

from syrupy.extensions.amber import AmberSnapshotExtension

DIFFERENT_DIRECTORY = "__snaps_example__"


class DifferentDirectoryExtension(AmberSnapshotExtension):
    @property
    def _dirname(self) -> str:
        return str(
            Path(self.test_location.filepath).parent.joinpath(DIFFERENT_DIRECTORY)
        )


@pytest.fixture
def snapshot(snapshot):
    return snapshot.use_extension(DifferentDirectoryExtension)


def test_case_1(snapshot):
    assert "Syrupy is amazing!" == snapshot



"""
Example: Custom Snapshot Name
"""
from typing import Union

import pytest

from syrupy.extensions.amber import AmberSnapshotExtension


class CanadianNameExtension(AmberSnapshotExtension):
    def get_snapshot_name(self, *, index: Union[str, int]) -> str:
        original_name = super(CanadianNameExtension, self).get_snapshot_name(
            index=index
        )
        return f"{original_name}🇨🇦"


@pytest.fixture
def snapshot(snapshot):
    return snapshot.use_extension(CanadianNameExtension)


def test_canadian_name(snapshot):
    assert "Name should be test_canadian_name🇨🇦." == snapshot



from invoke import task

from .utils import ctx_run


@task(default=True)
def dev(ctx):
    """
    Install the current development version of syrupy
    """
    ctx_run(ctx, "python -m pip install -U .")
